# -*- coding: utf-8 -*-
"""
Generate a roche-otss licensing file.

OTSS will call dpkg -l (``dpkg(1)``) and read the license of all packages,
match them to known licenses.
"""

import codecs
import datetime
import functools
import getpass
import logging
import os.path
import subprocess
import sys
import time
from multiprocessing import Pool
from os.path import join as pjoin

import click
import sh
import six

from . import cmd
from .common import prepare_text_for_xml
from .docx import Docx
from .fuzzymatch import close_handles
from .log import lg, log_to_stderr
from .package import Package, SkipException
from .workdir import download_licenses, prepare_workdir, read_licenses

show_man = functools.partial(cmd.show_man, "otss-gen")
package_dir = cmd.package_dir
if six.PY2:
    XFileNotFoundError = IOError  # noqa
else:
    XFileNotFoundError = FileNotFoundError  # noqa


def get_subject():
    """Return the subject of the document by reading /etc/os-release."""
    return subprocess.check_output([
        "bash",
        "-c",
        "source /etc/os-release; echo $PRETTY_NAME"
    ]).decode("UTF-8").strip()


def parse_package_list(input_stream, package_ctor):
    """Return a list of Package objects from the given input stream.

    Parse the package information from the given input stream, which
    should be the output of ``dpkg -l``. The result is a list of PkgDef
    named tuples.
    """
    out_list = []

    for line in input_stream:
        line = line.strip()
        if line.startswith("ii"):
            pkg = package_ctor(line)
            if pkg.blacklisted:
                continue

            out_list.append(pkg)
    return out_list


_config = {
    'blacklist':  pjoin(package_dir, "blacklist.list"),
    'overrides':  None,
    'licenses': (
        "https://github.com/spdx/license-list/archive/master.tar.gz"
    ),
    'debug':      False,
    'subject':    get_subject(),
    'version':    "draft",
    'sapdok-nr':  "1234NR",
    'sapdok-vr':  "1234VR",
    'sapdok-ktl': "1234KTL",
    'sapdok-ar':  "1234AR",
    'builddrop':  "path/to/builddrop",
}


@click.command(context_settings=cmd.CONTEXT_SETTINGS)
@click.argument(
    'dest',
    nargs=1,
    default=lambda: _config.get('dest', "out"),
    type=click.Path(
        exists=True,
        file_okay=False,
        writable=True,
        allow_dash=False
    ),
)
@click.option(
    '-i',
    '--input',
    'input_',
    nargs=1,
    default=lambda: _config.get('input'),
    help="Input file, if not defined dpkg -l will be called",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        allow_dash=True
    ),
)
@click.option(
    '--oldversion',
    'oldversion',
    nargs=1,
    default=lambda: _config.get('oldversion'),
    help="Optional old vesion of the dpkg list file, used for --delta modes",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        allow_dash=True
    ),
)
@click.option(
    '--delta',
    'delta',
    nargs=1,
    default=lambda: _config.get('delta'),
    help="Delta document mode. Accepts either 'document' or 'table'.",
    type=click.Choice(['document', 'table'])
)
@click.option(
    '-b',
    '--blacklist',
    nargs=1,
    default=lambda: _config.get('blacklist'),
    help="Blacklist file, if not defined default blacklist will be used",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
    ),
)
@click.option(
    '-l',
    '--licenses',
    nargs=1,
    default=lambda: _config.get('licenses'),
    help="Directory or url containing licenses. Defaults to: "
         "https://github.com/spdx/license-list/archive/master.tar.gz",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
    ),
)
@click.option(
    '-c',
    '--config',
    nargs=1,
    help="Config file, can contain any option",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
    ),
)
@click.option(
    '-s',
    '--subject',
    nargs=1,
    default=lambda: _config.get('subject'),
    help="Subject of the document, default '%s'" % get_subject(),
    type=str
)
@click.option(
    '-dv',
    '--doc-version',
    nargs=1,
    default=lambda: _config.get('version'),
    help="Version of the document, default 'draft'",
    type=str
)
@click.option(
    '--sapdok-nr',
    nargs=1,
    default=lambda: _config.get('sapdok-nr'),
    help="SAP Dok NR",
    type=str
)
@click.option(
    '--sapdok-ar',
    nargs=1,
    default=lambda: _config.get('sapdok-ar'),
    help="SAP Dok AR",
    type=str
)
@click.option(
    '--sapdok-ktl',
    nargs=1,
    default=lambda: _config.get('sapdok-ktl'),
    help="SAP Dok KTL",
    type=str
)
@click.option(
    '--sapdok-vr',
    nargs=1,
    default=lambda: _config.get('sapdok-vr'),
    help="SAP Dok VR",
    type=str
)
@click.option(
    '--builddrop',
    nargs=1,
    default=lambda: _config.get('builddrop'),
    help="Path to builddrop",
    type=str
)
@click.option(
    '--debug/--no-debug',
    '-d/',
    default=lambda: _config.get('debug'),
    help="Log debug output to stderr",
)
@click.option(
    '--help',
    '-h/',
    help="""Show otss-gen the man page.""",
    is_flag=True,
    callback=show_man,
    expose_value=False,
    is_eager=True
)
def cli(
        dest,
        input_,
        oldversion,
        delta,
        config,
        blacklist,
        licenses,
        subject,
        doc_version,
        sapdok_nr,
        sapdok_ar,
        sapdok_ktl,
        sapdok_vr,
        builddrop,
        debug,
):
    """
    Tool to generate the OTSS Documentation.

    DEST     Destination- and work-directory. Defaults to: out.
    """
    blacklist_data = read_blacklist(blacklist)
    if debug:
        log_to_stderr()
    else:
        log_to_stderr(logging.CRITICAL)

    if delta and not oldversion:
        print("Error: If you enable delta mode, you need to "
              "pass an --oldversion file")
        sys.exit(1)

    if oldversion and not delta:
        delta = 'table'  # default to table

    if input_ is None:
        lg.info("Reading packages from dpkg -l")
        stream = sh.dpkg("-l", _tty_out=False)
    elif os.path.isfile(input_):
        lg.info("Reading packages from %s", input_)
        stream = codecs.open(input_, "r", encoding="UTF-8")
    elif input_ == "-":
        lg.info("Reading packages from stdin")
        stream = sys.stdin
    else:
        raise ValueError("Cannot read input")
    prepare_workdir(dest)
    if licenses.startswith("http") or licenses.startswith("ftp"):
        for _ in range(4):
            try:
                licenses_path = download_licenses(licenses, dest)
                break
            except KeyError:
                time.sleep(1)
    else:
        if not os.path.exists(licenses):
            raise ValueError("Path to licenses does not exist")
        licenses_path = licenses
    (compare, licenses, names) = read_licenses(dest, licenses_path)

    # Constructor for generating Packages. this is used
    # both for parsing regular package lists, and lists for
    # old versions
    def package_ctor(line):
        return Package(
            dest, blacklist_data,
            compare, licenses,
            names, line, pool
        )

    pkgs         = []
    try:
        pool = Pool()
        read_pkgs = parse_package_list(stream, package_ctor)
        for pkg in read_pkgs:
            try:
                pkg.read()
                pkgs.append(pkg)
            except sh.ErrorReturnCode_100:
                lg.critical(
                    "Package not in repository, ignoring: %s",
                    pkg.name
                )
            except SkipException:
                pass
            except XFileNotFoundError:
                lg.critical(
                    "Package exists but has no copyright file, ignoring: %s",
                    pkg.name
                )
        if not (pkgs and licenses):
            raise ValueError("Please specify at least one package/license")
    finally:
        pool.terminate()
    close_handles()

    pkg_changes = []   # list of tuples with old and new pkg
    pkgs_changed = []  # list of packages which have changed
    if oldversion:
        old_pkgs = parse_package_list(open(oldversion, 'r'), package_ctor)
        pkgs_map = {p.name: p for p in pkgs}
        old_pkgs_map = {p.name: p for p in old_pkgs}
        all_pkg_names = sorted(pkgs_map.keys() | old_pkgs_map.keys())
        for pkg_name in all_pkg_names:
            old_pkg = old_pkgs_map.get(pkg_name)
            pkg = pkgs_map.get(pkg_name)

            if not old_pkg or not pkg or old_pkg.version != pkg.version:
                pkg_changes.append((old_pkg, pkg))
                if pkg:
                    pkgs_changed.append(pkg)

    # in delta document mode we need to reduce pkgs to packages changed
    pkgs = delta == 'document' and pkgs_changed or pkgs
    notes = {package.license_hash_lower: package for package in pkgs}
    notes = [
        package for package in sorted(notes.values(), key=lambda x: x.name)
    ]
    docx_licenses = prepare_docx_licenses(pkgs, licenses)
    docx = Docx(
        pjoin(package_dir, "basic"),
        pjoin(dest, "stage"),
        pjoin(dest, "output.docx"),
    )
    docx.stage_basic()
    docx.stage_templates({
        "creator": getpass.getuser(),
        "created": datetime.datetime.utcnow().strftime(
            "%Y-%m-%dT%H:%M:%S"
        ),
        "subject":       subject,
        "version":       doc_version,
        "sapdok_nr":     sapdok_nr,
        "sapdok_ar":     sapdok_ar,
        "sapdok_ktl":    sapdok_ktl,
        "sapdok_vr":     sapdok_vr,
        "builddrop":     builddrop,
        "licenses":      docx_licenses,
        "inventory":     pkgs,
        "changes":       pkg_changes,
        "notes":         notes,
    })
    docx.make_docx()


def prepare_docx_licenses(packages, licenses):
    """Prepare and filter the licenses to be contained in the docx."""
    licenses_set = {}
    for package in packages:
        for name, license in package.licenses:
            licenses_set[license[0].ldigest] = (name, license)
    licenses_sort = sorted(licenses_set.values(), key=lambda x: x[0])
    docx_licenses = [
        (
            name,
            license[0].ldigest,
            prepare_text_for_xml(license[0].text)
        )
        for name, license in licenses_sort
    ]
    return docx_licenses


def read_blacklist(blacklist):
    """Read blacklist file."""
    with codecs.open(blacklist, encoding="UTF-8") as f:
        return set(f.readline())

try:
    c = None
    for i, arg in enumerate(sys.argv):
        if arg in ("-c", "--config"):
            if six.PY2:
                import ConfigParser
                import StringIO
                c = ConfigParser.ConfigParser()
                with codecs.open(sys.argv[i + i], encoding="UTF-8") as f:
                    config = "[a]%s%s" % (
                        os.linesep,
                        f.read()
                    )
                c.readfp(StringIO.StringIO(config))
            else:
                import configparser
                c = configparser.ConfigParser()
                with codecs.open(sys.argv[i + i], encoding="UTF-8") as f:
                    config = "[a]%s%s" % (
                        os.linesep,
                        f.read()
                    )
                c.read_string(config)
            break
    if c:
        for section in c.sections():
            for option in c.options(section):
                _config[option] = c.get(section, option)
except Exception:
    raise ValueError("Could not read config file")
