"""Import and export otss cache."""

import functools
import glob
import os
import sys

import click

from . import cmd
from .fuzzymatch import close_handles, get_db

show_man = functools.partial(cmd.show_man, "otss-cache")


def db_res_gen(cursor, arraysize=1000):
    """A generator that uses fetchmany to keep memory usage down."""
    while True:
        results = cursor.fetchmany(arraysize)
        if not results:
            break
        for result in results:
            yield result


@click.command(context_settings=cmd.CONTEXT_SETTINGS)
@click.argument(
    'cache',
    nargs=1,
    default="cache",
    type=click.Path(
        exists=True,
        file_okay=False,
        writable=True,
        allow_dash=False
    ),
)
@click.argument(
    'dest',
    default="out",
    nargs=1,
    type=click.Path(
        exists=True,
        file_okay=False,
        writable=True,
        allow_dash=False
    ),
)
@click.option(
    '--import',
    '-i/',
    'import_',
    is_flag=True,
    help="Import the cache",
)
@click.option(
    '--export',
    '-e/',
    is_flag=True,
    help="Export the cache",
)
@click.option(
    '--help',
    '-h/',
    help="""Show otss-gen the man page.""",
    is_flag=True,
    callback=show_man,
    expose_value=False,
    is_eager=True
)
def cli(
        cache,
        dest,
        import_,
        export
):
    """
    Tool to import/export the cache to/from a working directory.

    It exports simple text-files which can be checked into git and easily
    merged and edited.

    CACHE    Directory containing the cache. Defaults to: cache.
    DEST     Destination- and work-directory. Defaults to: out.
    """
    dcache = os.path.join(dest, "cache")
    if not os.path.exists(dcache):
        os.mkdir(dcache)
    db = get_db(dcache)
    with db:
        if export:
            done = 0
            count = int(db.execute("""
                SELECT
                    count(key)
                FROM
                    ratio
            """).fetchone()[0])
            print("Writing %d records" % count)
            parts = set(db_res_gen(db.execute("""
                SELECT DISTINCT
                    part
                FROM
                    ratio
            """)))
            for part in parts:
                with open(os.path.join(cache, "%s.txt" % part), "w") as fpart:
                    lines = db_res_gen(db.execute("""
                        SELECT
                            key,
                            ratio
                        FROM
                            ratio
                        WHERE
                            part = ?
                        ORDER BY
                            key
                    """, part))
                    for line in lines:
                        done += 1
                        fpart.write("%s %s\n" % line)
                        if (done % 1000) == 0:
                            sys.stdout.write(
                                "\b\b\b\b\b\b%3.2f%%" % (
                                    float(done) / count * 100
                                )
                            )
                            sys.stdout.flush()
            print("\b\b\b\b\b\b\b100.00%%\nWrote %d records" % done)
            with open(os.path.join(cache, "lines.inf"), "w") as f:
                f.write(str(done))
        elif import_:
            done = 0
            with open(os.path.join(cache, "lines.inf"), "r") as f:
                count = int(f.readline())
            print("Reading %d records" % count)
            for part in glob.iglob(
                    os.path.join(cache, "*.txt")
            ):
                partnum = part.split(os.path.sep)[-1][:-4]
                with open(os.path.join(part), "r") as fpart:
                    for line in fpart:
                        done += 1
                        key, ratio = line.split(' ')
                        db.execute("""
                            INSERT OR REPLACE INTO
                                ratio
                            VALUES
                                (?, ?, ?)
                        """, (key, partnum, ratio))
                        if (done % 1000) == 0:
                            sys.stdout.write(
                                "\b\b\b\b\b\b%3.2f%%" % (
                                    float(done) / count * 100
                                )
                            )
                            sys.stdout.flush()
            print("\b\b\b\b\b\b\b100.00%%\nRead %d records" % done)
        else:
            print("Please use either the export or the import option.")
    print("Vaccum database")
    db.isolation_level = None
    db.execute("vacuum")
    close_handles()
