"""Create docx file."""

import codecs
import os
import shutil
import zipfile

import jinja2

from .log import lg


class Docx(object):
    """
    Generate docx file from basic template, jinja templates and license input.

    :param str         basic: Path to basic template
    :param str         stage: Path to staging directory
    :param str           out: Path to docx file
    """

    def __init__(self, basic, stage, out):
        self.basic        = basic
        self.stage        = stage
        self.out          = out
        self.env          = jinja2.Environment(
            loader=jinja2.PackageLoader(
                'otss', 'templates'
            )
        )

    def stage_basic(self):
        shutil.rmtree(self.stage)
        shutil.copytree(self.basic, self.stage)
        os.rename(
            os.path.join(self.stage, "_Content_Types_.xml"),
            os.path.join(self.stage, "[Content_Types].xml"),
        )

    def stage_templates(self, context):
        self.render_template(
            "docProps/custom.xml",
            context
        )
        self.render_template(
            "docProps/core.xml",
            context
        )
        self.render_template(
            "word/header2.xml",
            context
        )
        self.render_template(
            "word/header3.xml",
            context
        )
        self.render_template(
            "word/document.xml",
            context
        )

    def render_template(self, template, context):
        lg.debug("Render template: %s", template)
        path_parts = template.split("/")
        tmpl = self.env.get_template(
            "%s.jinja" % (
                os.path.join(*path_parts)
            )
        )
        with codecs.open(os.path.join(
                self.stage,
                *path_parts
        ), "w", "UTF-8") as f:
            f.write(tmpl.render(context))

    def make_docx(self):
        with zipfile.ZipFile(
                self.out,
                mode="w",
                compression=zipfile.ZIP_DEFLATED
        ) as f:
            lg.debug("Write zip %s", self.out)
            stage_len = len(self.stage) + 1
            for root, _, files in os.walk(self.stage):
                dir_path = root[stage_len:]
                for file_ in files:
                    file_path = os.path.join(root, file_)
                    zip_path = os.path.join(dir_path, file_)
                    lg.debug(
                        "Append file %s as %s to zip",
                        file_path,
                        zip_path
                    )
                    f.write(file_path, zip_path)
