# -*- coding: utf-8 -*-
"""Represents a package."""

import codecs
import email
import hashlib
import os.path
import re
from os.path import abspath, dirname

import sh

from .common import flatten_config, prepare_text_for_xml
from .fuzzymatch import partial_ratio, set_ratio
from .log import lg

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


class SkipException(Exception):
    """Package should be skipped."""

    pass

_license_match = re.compile("common-licenses/([^\s\"',;\’\)]+).*")
package_dir = abspath(dirname(__file__))

_override_template = """[Package]
; licenses: comma separated names of licenses we know this package has
licenses = %s

; match: True | False. Set to False if no automatic match should be performed
match    = %s"""

_common_licenses_aliases = {
    "expat":           "mit",
}

_match_aliases = {
    "lgpl-2.1-short0":        "lgpl-2.1",
    "lgpl-2.1-short1":        "lgpl-2.1",
    "lgpl-2.1-libc":          "lgpl-2.1+",
    "lgpl-2.1-stupid-short0": "lgpl-2.1+",
    "gpl-2-linux":            "gpl-2+",
    "gpl-2-short0":           "gpl-2+",
    "gpl-2-short1":           "gpl-2+",
    "gpl-2-short2":           "gpl-2+",
    "gpl-2-short3":           "gpl-2+",
    "gpl-2-short4":           "gpl-2+",
    "gpl-2-short5":           "gpl-2+",
    "gpl-2-short6":           "gpl-2+",
    "gpl-2-short7":           "gpl-2+",
    "gpl-2-stupid-short0":    "gpl-2+",
    "gpl-2-stupid-short1":    "gpl-2+",
    "gpl-3-short0":           "gpl-3+",
    "gpl-3-short1":           "gpl-3+",
    "gpl-3-short2":           "gpl-3+",
    "gpl-3-bash":             "gpl-3+",
    "mit-sipb0":              "mit",
    "mit-sipb1":              "mit",
    "mit-xorg":               "mit",
    "mit-short":              "mit",
    "mit-bell":               "mit",
    "bsd-x0":                 "bsd",
    "keith-mit":              "mit",
    "expat":                  "mit",
    "bsd-2-clause-variant":   "bsd-2-clause",
    "bsd-pam":                "bsd-3-clause",
    "bsd-pcre":               "bsd-3-clause",
    "bsd-venema":             "bsd",
    "isc-short":              "isc",
    "x11-short":              "x11",
}

_clean_license = re.compile("[`'*\/\\\"!?,.;()|\s-]")


def _do_ratio(arg):
    """Do levensthein distance in parallel."""
    (
        cache,
        package_name,
        compare_name,
        license_dgst,
        compare_dgst,
        license,
        compare,
    ) = arg
    ratio = partial_ratio(
        cache,
        (license_dgst, compare_dgst),
        _clean_license.sub('', license),
        _clean_license.sub('', compare)
    )
    lg.debug(
        "Check %s for match against %s (%02d)",
        package_name,
        compare_name,
        ratio
    )
    return (license_dgst, ratio)


class Package(object):
    """
    Parses multiple sources: dpkg list, custom .ini file and .spdx file.

    The .ini values will never be overwritten by parsed content.

    :param             dest: Destination
    :param        blacklist: Packages that are ignored
    :param compare_licenses: Licenses to compare against
    :param   known_licenses: Known licenses by spdx
    :param    license_names: Known licenses longest first
    :param   dpkg_list_line: The raw dpkg list entry
    :param             pool: Multiprocessing pool
    """

    _non_std = 0

    def __init__(
            self,
            dest,
            blacklist,
            compare_licenses,
            known_licenses,
            license_names,
            dpkg_list_line,
            pool
    ):
        self.copyright           = []
        self.licenses            = set()
        self.license_definitions = []
        self.license_text        = None
        self.license_text_lower  = None
        self.license_hash_lower  = None
        self.license_text_tmpl   = None
        self.note                = None
        self.match               = False
        self._pool               = pool
        self._blacklist          = blacklist
        self._compare_licenses   = compare_licenses
        self._known_licenses     = known_licenses
        self._license_name       = license_names
        self._dest               = dest
        self._unknown            = os.path.join(dest, "unknown")
        self._override           = os.path.join(dest, "override")
        self._cache              = os.path.join(dest, "cache")
        (
            self.name,
            self.version,
            self.architecture,
            self.description
        ) = dpkg_list_line.split(None, 4)[1:]  # Stupid old python on wheezy
        # Strip architecture identifier from name
        if ':' in self.name:
            self.name = self.name[0:self.name.rfind(':')]

        self.license_file = os.path.join(
            os.path.sep,
            "usr",
            "share",
            "doc",
            self.name,
            "copyright"
        )

    def read(self):
        """Read all package information."""
        standard = False
        self.read_control()
        self.read_license()
        if not self.license_text:
            lg.critical("Package has no license-file: %s", self.name)
            raise SkipException()
        self.override()

        if self.licenses:
            standard = True
            lics = set(self.licenses)
            self.match_license_names()

            if not self.licenses:
                lg.critical(
                    "Package defined %s licenses but no match found: %s",
                    lics,
                    self.name
                )

        if (not self.licenses) or self.match:
            if not (self.licenses or standard):
                Package._non_std += 1
                lg.debug(
                    "Copyright file of %s is non-stanard (%d), "
                    "matching licenses",
                    self.name,
                    Package._non_std
                )
            self.match_licenses()
            self.clean_versions()
            self.grep_licenses()
        if self.licenses:
            self.clean_update_versions()
            self.replace_license_aliases()
            lg.debug(
                "Package %s has licenses: %s",
                self.name,
                sorted(self.licenses)
            )
            self.licenses = [
                (x, self._known_licenses[x]) for x in sorted(self.licenses)
            ]
        else:
            lg.critical(
                "Package %s: no licenses detected",
                self.name,
            )

        self.note = prepare_text_for_xml(self.license_text)

    def match_license_names(self):
        """Filter licenses list to contain known licenses."""
        lg.debug(
            "Filtering %s licenses: %s",
            self.licenses,
            self.name,
        )
        to_remove = []
        for license in self.licenses:
            if license not in self._known_licenses:
                to_remove.append(license)
        for license in to_remove:
            self.licenses.remove(license)
            if (
                    "public domain" in license or
                    "public-domain" in license or
                    "publicdomain" in license
            ):
                self.licenses.add("public-domain")
            # Re-add the longest partial match of a known license
            for compare in self._license_name:
                if compare in license:
                    ignore = False
                    for existing in self.licenses:
                        if compare in existing:
                            ignore = True
                            break
                    if not ignore:
                        self.licenses.add(compare)

    def clean_versions(self):
        """If we have multiple versions check for 'version x'."""
        flat = None
        if self.licenses.issuperset({"gpl-2+", "gpl-3+"}):
            flat = _clean_license.sub('', self.license_text_lower)
            if "version3" not in flat:
                self.licenses.remove("gpl-3+")
            if "version2" not in flat:
                self.licenses.remove("gpl-2+")
        if self.licenses.issuperset({"gpl-2", "gpl-3"}):
            if not flat:
                flat = _clean_license.sub('', self.license_text_lower)
            if "version3" not in flat:
                self.licenses.remove("gpl-3")
            if "version2" not in flat:
                self.licenses.remove("gpl-2")
        found = set()
        for lic in self.licenses:
            if "lgpl" in lic:
                found.add(lic)
        if found:
            if not flat:
                flat = _clean_license.sub('', self.license_text_lower)
            if not ("lgpl" in flat or "lesser" in flat):
                self.licenses -= found

    def clean_update_versions(self):
        """If we versions that are updates."""
        if "gpl-3+" in self.licenses:
            self.licenses.discard("gpl-3")
            self.licenses.discard("gpl")
        if "gpl-2+" in self.licenses:
            self.licenses.discard("gpl-2")
            self.licenses.discard("gpl")
        if "lgpl-2+" in self.licenses:
            self.licenses.discard("lgpl-2")

    def replace_license_aliases(self):
        """Replace licenses with common license names."""
        for license in self.licenses:
            if license in _common_licenses_aliases:
                self.licenses.remove(license)
                self.licenses.add(_common_licenses_aliases[license])

    def grep_licenses(self):
        """Find licenses using a regex.

        This is the least reliable method, therefor it is applied last.
        """
        match = _license_match.finditer(self.license_text_lower)
        if match:
            for lic in match:
                lic_name = lic.group(1).strip('.').lower()
                if lic_name == "gpl":
                    lic_name = "gpl-3"
                self.licenses.add(lic_name)

    def match_licenses(self):
        """Match license text against loaded licenses."""
        margin = 15
        found = 0
        test_ratio = 103
        while found < margin and test_ratio > 0:
            to_check = []
            test_ratio -= 3
            for digest in self._compare_licenses.keys():
                license = self._compare_licenses[digest]
                ratio = set_ratio(
                    self._cache,
                    (self.license_hash_lower, digest),
                    self.license_text_lower,
                    license[0].text.lower()
                )
                if ratio >= test_ratio:
                    to_check.append(digest)
            found = len(to_check)

        check_set = set(to_check)
        lg.debug(
            "Prefiltered %d of %d licenses",
            len(check_set),
            len(self._known_licenses)
        )

        ratio_map = []
        for digest in check_set:
            license = self._compare_licenses[digest]
            ratio_map.append((
                self._cache,
                self.name,
                license[0].name,
                digest,
                self.license_hash_lower,
                # limit package license to 40k to avoid out of memory
                self.license_text_lower[:2 ** 10 * 40],
                # limit known licenses to 8k to avoid out of memory
                license[0].text[:2 ** 10 * 8].lower()
            ))
        ratio_map = self._pool.imap(_do_ratio, ratio_map)
        for ratio_entry in ratio_map:
            license = self._compare_licenses[ratio_entry[0]]
            ratio   = ratio_entry[1]
            if ratio > 95:
                for license_copy in license:
                    name = license_copy.name
                    if name in _match_aliases:
                        self.licenses.add(_match_aliases[name])
                    else:
                        self.licenses.add(name)

    @property
    def blacklisted(self):
        return self.name in self._blacklist

    def fill_license_fields(self, header):
        for key in header.keys():
            attr = key.replace("-", "_").lower()
            value = header[key].strip()
            if attr == "copyright":
                if not self.is_debian:
                    self.copyright.append(value)
            elif attr == "files":
                self.is_definition = False
                self.is_debian = value.startswith("debian/")
            elif attr == "license":
                if not self.is_debian:
                    if self.is_definition:
                        self.license_definitions.append(value)
                    else:
                        if os.linesep in value:
                            self.license_definitions.append(value)
                            lic_line = value.split(os.linesep)[0].strip()
                            if lic_line and len(lic_line) < 60:
                                self.licenses.add(lic_line.lower())
                        else:
                            self.licenses.add(value.lower())
            else:
                while hasattr(self, attr):
                    attr = "%s_" % attr
                setattr(self, attr, value)

    def read_license(self):
        """Read the license file of the package."""
        if os.path.exists(self.license_file):
            parser = email.parser.FeedParser()
            self.is_debian = False
            self.is_definition = True
            with codecs.open(self.license_file, encoding="UTF-8") as f:
                for line in f:
                    if line.strip(os.linesep) == "":
                        self.is_definition = True
                        self.fill_license_fields(parser.close())
                        parser = email.parser.FeedParser()
                    else:
                        parser.feed(line)
            self.fill_license_fields(parser.close())
            with codecs.open(self.license_file, encoding="UTF-8") as f:
                self.license_text = f.read()
                self.license_text_lower = self.license_text.lower()
                self.license_hash_lower = hashlib.md5(
                    self.license_text_lower.encode("utf-8")
                ).hexdigest()

    def read_control(self):
        """Read control file of package."""
        control = sh.apt_cache(["show", self.name], _tty_out=False)
        parser = email.parser.FeedParser()
        for line in control:
            parser.feed(line)
        header = parser.close()
        for key in header.keys():
            attr = key.replace("-", "_").lower()
            while hasattr(self, attr):
                attr = "%s_" % attr
            setattr(self, attr, header[key].strip())

    def override(self):
        """Parse the custom .ini values for this package, if there is one.

        Those values have the highest relevance. All parsed content will be
        overwritten by those values.
        """
        filenames = []
        filenames.append(os.path.join(
            package_dir,
            "override",
            "%s.ini" % self.name
        ))
        filenames.append(os.path.join(
            self._override,
            "%s.ini" % self.name
        ))
        for filename in filenames:
            if os.path.exists(filename):
                config = configparser.ConfigParser()
                config.read(filename)
                config_flat = flatten_config(config)
                lg.info("Parsed override (.ini) for %s", self.name)
                for key in config_flat:
                    value = config_flat[key]
                    if key == 'licenses':
                        value = [v.strip().lower() for v in value.split(',')]
                    elif value.lower() == "true":
                        value = True
                    elif value.lower() == "false":
                        value = False
                    else:
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                    setattr(self, key, value)
                    lg.debug(
                        "Override %s: %s with %s",
                        self.name,
                        key,
                        config_flat[key],
                    )
