# -*- coding: utf-8 -*-
"""Manage the workdir."""

import codecs
import glob
import hashlib
import os.path
import tarfile
from collections import defaultdict, namedtuple
from os.path import abspath, dirname

import requests
import six
from chardet.universaldetector import UniversalDetector

from .log import lg

package_dir = abspath(dirname(__file__))

if six.PY2:
    FileExistsError = OSError

_licenses_blacklist = {}  # A set not a dict

License = namedtuple('License', ['text', 'name', 'ldigest', 'udigest'])


def prepare_workdir(dest):
    """Prepare folder-structure in workdir.

    :param dest: Directory to prepare
    """
    for dir_ in ("cache", "override", "license", "stage"):
        try:
            os.mkdir(os.path.join(dest, dir_))
        except FileExistsError:
            pass


def download_licenses(url, dest):
    """Download the licenses into the workdir.

    :param  url: Url to download from
    :param dest: Directory to unpack into
    """
    local_filename = os.path.join(dest, "license.tar.gz")
    r = requests.get(url, stream=True)
    size = int(r.headers['Content-Length'])
    downloaded = 0
    with open(local_filename, 'wb') as f:
        for chunk in r.iter_content(chunk_size=4096):
            if chunk:
                f.write(chunk)
                downloaded += 4096
                lg.debug("Downloaded %02d%%", int(downloaded * 100 / size))
    tar = tarfile.open(local_filename)
    tar.extractall(path=dest)
    path = tar.getmembers()[0].path.split(os.path.sep)[0]
    tar.close()
    path = os.path.join(dest, path)
    if not os.path.exists(path):
        raise OSError("After extracting path to licenses does not exist")
    return path


def read_licenses(dest, directory):
    """Read licenses from directory.

    :param directory: Containing the licenses
    :rtype:           dict
    """
    licenses = defaultdict(list)
    license_names = dict()
    pattern_workdir = os.path.join(directory, "*.txt")
    pattern_license = os.path.join(dest, "license", "*.txt")
    pattern_package = os.path.join(
        package_dir,
        "license",
        "*.txt"
    )
    files = []
    files.extend(glob.iglob(pattern_workdir))
    files.extend(glob.iglob(pattern_license))
    files.extend(glob.iglob(pattern_package))
    for license in files:
        file_ = license.split(os.path.sep)[-1]
        lg.debug("Reading license %s" % file_)
        name  = os.path.extsep.join(
            file_.split(os.path.extsep)[:-1]
        ).lower()
        if name.startswith("deprecated_"):
            name = name[11:]
        if name in _licenses_blacklist:
            continue
        if ' ' in name:
            continue
        text = None
        try:
            with codecs.open(license, encoding="UTF-8") as f:
                text = f.read()
        except UnicodeDecodeError:
            with open(license, "rb") as f:
                detector = UniversalDetector()
                for line in f.readlines():
                    detector.feed(line)
                    if detector.done:
                        break
                detector.close()
                encoding = detector.result['encoding']
            if encoding:
                with codecs.open(license, encoding=encoding) as f:
                    text = f.read()
        udigest = hashlib.md5(text.encode('UTF-8')).hexdigest()
        ldigest = hashlib.md5(text.lower().encode('UTF-8')).hexdigest()
        licenses[ldigest].append(License(text, name, ldigest, udigest))
    for license in list(licenses.values()):
        for instance in license:
            license_names[instance.name] = license
    license_names_longest = sorted(license_names.keys(), key=len, reverse=True)
    return (licenses, license_names, license_names_longest)
