"""Do fuzzy matching witch cache."""

import atexit
import functools
import hashlib
import os
import sqlite3

from fuzzywuzzy import fuzz

from .log import lg

_db = None
_cache = {}


def close_handles():
    """Close the dbm and the protocol."""
    global _db
    if _db:
        lg.debug("Closing sqlite")
        _db.close()
        _db = None

atexit.register(close_handles)


def get_db(cache):
    """Open a db."""
    global _db
    global _proto
    if not _db:
        lg.debug("Opening sqlite")
        path = os.path.join(cache, "ratio.db")
        path_exists = os.path.exists(path)
        _db = sqlite3.connect(path)
        if not path_exists:
            with _db:
                _db.execute("""
                    CREATE TABLE
                        ratio
                    (
                        key TEXT PRIMARY KEY,
                        part TEXT,
                        ratio INTEGER
                    )
                """)
                _db.execute("""
                    CREATE INDEX idx_part ON ratio(part);
                """)
    return _db


def ratio(func, ext, cache, k, x, y):
    """
    Fuzzywuzzy ratio with cache.

    :param func:  ratio function
    :param ext:   cache file extension
    :param cache: cache directory
    :param k:     key to look up
    :param x:     string to match
    :param y:     string to match
    :rtype:       integer
    """
    key   = hashlib.sha256(str((k, ext)).encode()).hexdigest()
    ratio = _cache.get(key)
    if ratio:
        return _cache[key]
    db = get_db(cache)
    with db:
        res = db.execute("""
            SELECT
                ratio
            FROM
                ratio
            WHERE
                key == ?
            """, (key,)).fetchone()
    if res:
        return int(res[0])
    else:
        ratio = func(
            x,
            y,
        )
        _cache[key] = ratio
        part = key[:2]
        with db:
            db.execute("""
                INSERT INTO
                    ratio
                VALUES
                    (?, ?, ?)
            """, (key, part, ratio))
        return ratio

partial_ratio = functools.partial(ratio, fuzz.partial_ratio, "part")
set_ratio = functools.partial(ratio, fuzz.token_set_ratio, "set")
sort_ratio = functools.partial(
    ratio, fuzz.token_sort_ratio, "sort"
)
